import json
import os
import string
import warnings
import re
from langdetect import detect_langs
from langdetect.lang_detect_exception import LangDetectException
from lingua import LanguageDetectorBuilder
import random
from termcolor import colored

# Unicode ranges for various writing systems
WRITING_SYSTEMS_UNICODE_RANGES = {
    'zh': [(0x4E00, 0x9FFF), (0x3400, 0x4DBF), (0x20000, 0x2A6DF), (0x2A700, 0x2B73F), (0x2B740, 0x2B81F)],  # Chinese
    'ja': [(0x3040, 0x309F), (0x30A0, 0x30FF)],  # Japanese (Hiragana and Katakana)
    'ko': [(0xAC00, 0xD7AF)],  # Korean (Hangul)
    'ar': [(0x0600, 0x06FF), (0x0750, 0x077F), (0x08A0, 0x08FF)],  # Arabic
    'cy': [(0x0400, 0x04FF)],  # Cyrillic
    'deva': [(0x0900, 0x097F)],  # Devanagari (used for Hindi, Sanskrit, etc.)
    'he': [(0x0590, 0x05FF)],  # Hebrew
    'th': [(0x0E00, 0x0E7F)],  # Thai
}

# Mapping of predefined language codes to specific colors
LANGUAGE_COLORS = {
    'en': 'green',
    'zh': 'yellow',
    'ja': 'cyan',
    'ko': 'blue',
    'ar': 'green',
    'cy': 'magenta',
    'hi': 'red',
    'mr': 'red',
    'he': 'white',
    'th': 'blue',
    'phoneme': 'blue',
    '??': 'red'
}

unknown_language_colors = {}


def random_color():
    colors = ['red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white']
    return random.choice(colors)


def print_colored_text(text):
    pattern = r'<(\w+)>(.*?)</\1>'
    last_pos = 0

    for match in re.finditer(pattern, text):
        lang, content = match.groups()
        start, end = match.span()

        if last_pos < start:
            print(text[last_pos:start], end="")

        if lang not in LANGUAGE_COLORS:
            if lang not in unknown_language_colors:
                unknown_language_colors[lang] = random_color()
            color = unknown_language_colors[lang]
        else:
            color = LANGUAGE_COLORS[lang]

        if lang == "??":
            color = 'red'

        print(colored(content, color), end="")
        last_pos = end

    if last_pos < len(text):
        print(text[last_pos:])


def load_manual_word_dict(file_path='manual_word_dict.json'):
    base_path = os.path.dirname(__file__)  # Directory of the current file
    file_path = os.path.join(base_path, file_path)
    try:
        with open(file_path, 'r') as f:
            return json.load(f)
    except FileNotFoundError:
        # print(f"Warning: {file_path} not found. Using empty dictionary.")
        return {}


def manual_tag(sentence, manual_word_dict):
    # Convert the sentence to lowercase and strip it of extra spaces
    sentence_lower = sentence.lower().strip()
    # Split the sentence into words
    words = sentence_lower.split()
    # Check each word against the manual dictionary
    for word in words:
        # Check if the word is in the manual dictionary
        if word in manual_word_dict:
            return manual_word_dict[word]
    return None  # Return None if no match is found


class Tokenizer:
    def __init__(self):
        self.min_confidence = 0.5
        self.manual_word_dict = load_manual_word_dict()
        self.detector = LanguageDetectorBuilder.from_all_languages().build()

    def detect_language(self, text):
        # Adjusted logic to improve language detection
        text_lower = text.lower().strip()
        manual_lang = manual_tag(text_lower, self.manual_word_dict)
        if manual_lang:
            return manual_lang
        try:
            langs = self.detector.detect_language_of(text)
            if langs is not None:
                langs = langs.iso_code_639_1.name.lower()
                return langs
            else:
                return '??'
        except LangDetectException:
            return '??'

    def is_writing_system(self, char, system):
        if len(char) > 1:
            # Valid punctuation characters, including space
            return all(self.is_writing_system(c) for c in char)  # Check each character individually
        else:
            code_point = ord(char)
            return any(start <= code_point <= end for start, end in WRITING_SYSTEMS_UNICODE_RANGES.get(system, []))

    def detect_japanese_korean_chinese(self, text):
        is_japanese = any(self.is_writing_system(char, 'ja') for char in text)
        is_korean = any(self.is_writing_system(char, 'ko') for char in text)
        is_chinese = any(self.is_writing_system(char, 'zh') for char in text)

        if is_japanese:
            return "ja"
        elif is_korean:
            return "ko"
        elif is_chinese:
            return "zh"
        else:
            return "??"

    def detect_writing_system(self, text):
        for system, ranges in WRITING_SYSTEMS_UNICODE_RANGES.items():
            if any(ord(char) in range(start, end + 1) for char in text for start, end in ranges):
                if system not in ["zh", "ja", "ko"]:
                    return system
                else:
                    return "cjk"
        return None

    def is_punctuation(self, char):
        not_punctuation = ["'", '"', "(", ")", "{", "}", "[", "]", "&"]
        if len(char) > 1:
            # Valid punctuation characters, including space
            return all(self.is_punctuation(c) for c in char)  # Check each character individually
        else:
            # Single character check (as per your original logic)
            return (not char.isalnum()  # Is not alphanumeric
                    and not char.isspace()  # Is not whitespace
                    and not self.is_writing_system(char, self.detect_writing_system(char))
                    and char not in not_punctuation)  # Is not in a writing system                                                                             self.detect_writing_system(
                                                                                                

    def split_text_by_writing_system(self, text):
        segments = []
        current_segment = ""
        current_type = None

        for char in text:
            if self.is_punctuation(char):
                if current_segment:
                    segments.append((current_segment, current_type))
                    current_segment = ""
                segments.append((char, "punctuation"))
                current_type = None
            else:
                char_system = self.detect_writing_system(char)
                if char_system != current_type:
                    if current_segment:
                        segments.append((current_segment, current_type))
                    current_type = char_system
                    current_segment = char
                else:
                    current_segment += char

        if current_segment:
            segments.append((current_segment, current_type))

        return segments

    @staticmethod
    def split_non_cjk_in_segment(text):
        return re.findall(r'[\u4E00-\u9FFF\u3400-\u4DBF\uF900-\uFAFF\u3040-\u309F\u30A0-\u30FF\uAC00-\uD7AF。]'
                          r'+(?:\s*)|[\w.,!?;:\'"(){}\[\]\-–—\s]+', text)

    def _tokenize(self, text):
        segments = self.split_text_by_writing_system(text)
        processed_segments = []

        for segment, seg_type in segments:
            if seg_type == "cjk":
                lang = self.detect_japanese_korean_chinese(segment)
                processed_segments.append(f"<{lang}>{segment}</{lang}>")
            elif seg_type in WRITING_SYSTEMS_UNICODE_RANGES:
                if seg_type != "deva":
                    processed_segments.append(f"<{seg_type}>{segment}</{seg_type}>")
                else:
                    lang = self.detect_language(segment)
                    processed_segments.append(f"<{lang}>{segment.strip()}</{lang}>")
            elif seg_type == "punctuation":
                processed_segments.append(f"<punctuation>{segment}</punctuation>")
            else:
                words = self.split_non_cjk_in_segment(segment)
                current_lang = None
                current_segment = ""

                for word in words:
                    if self.is_punctuation(word):
                        if current_segment:
                            processed_segments.append(f"<{current_lang}>{current_segment.strip()}</{current_lang}>")
                            current_segment = ""
                        processed_segments.append(f"<punctuation>{word}</punctuation>")
                        current_lang = None
                    else:
                        lang = self.detect_language(word)
                        if lang != current_lang:
                            if current_segment:
                                processed_segments.append(
                                    f"<{current_lang}>{current_segment.strip()}</{current_lang}>")
                                current_segment = ""
                            current_lang = lang
                        current_segment += word + " "

                # Handle any remaining text
                if current_segment:
                    processed_segments.append(f"<{current_lang}>{current_segment.strip()}</{current_lang}>")

        return "".join(processed_segments)

    def _group_segments(self, text):
        pattern = r'(<(\w+)>.*?</\2>|<punctuation>.*?</punctuation>)'
        tokens = re.findall(pattern, text)

        grouped_segments = []
        current_lang = None
        current_content = []

        for segment, lang in tokens:
            if lang:
                content = re.search(r'<\w+>(.*?)</\w+>', segment).group(1)
                if lang == current_lang:
                    current_content.append(content)
                else:
                    if current_content:
                        grouped_segments.append(f"<{current_lang}>{''.join(current_content)}</{current_lang}>")
                    current_lang = lang
                    current_content = [content]
            else:
                if current_content:
                    current_content[-1] += segment
                else:
                    grouped_segments.append(segment)

        if current_content:
            grouped_segments.append(f"<{current_lang}>{''.join(current_content)}</{current_lang}>")

        # Clean up punctuation handling
        return ''.join(grouped_segments).replace("<punctuation>", "").replace("</punctuation>", " ")

    def tokenize(self, text, group=True):
        # Split the input text into segments based on existing tags
        pattern = r'(<\w+>.*?</\w+>)|([^<]+)'  # Matches either tagged segments or untagged text
        segments = re.findall(pattern, text)

        processed_segments = []

        for tagged_segment, untagged_segment in segments:
            if tagged_segment:  # If this segment is already tagged, just add it
                processed_segments.append(tagged_segment)
            else:  # If the segment is untagged, process it as usual
                result = self._tokenize(untagged_segment)
                processed_segments.append(result)

        result = ''.join(processed_segments)

        if group:
            result = self._group_segments(result)

        if "<??>" in result:
            warnings.warn(
                "Your output contains tokenization errors. We were unable to detect a language or writing system, or there was an error in processing.")

        return result


# Main function
if __name__ == "__main__":
    input_text = "don't do that please"
    token = Tokenizer()
    processed_text = token.tokenize(input_text)
    print("Input text:")
    print(input_text)
    print("\nProcessed text:")
    print(processed_text)
    print_colored_text(processed_text)



import cutlet
import re

formal_to_informal = {

    'ワタクシ': 'わたし',
    'チカコ': 'しゅうこ',
    "タノヒト": "ほかのひと",

    # Add more mappings as needed
}

formal_to_informal2 = {

    "たのひと": "ほかのひと",
    "すうは": "かずは",

    # Add more mappings as needed
}

formal_to_informal3 = {

    "%": "％",
    "@": "あっとさいん",
    "$": "どる",
    "#": "はっしゅたぐ",
    "＄": "どる",
    "＃": "はっしゅたぐ",
    "何が": "なにが",

    "何も": "なにも",
    "何か": "なにか",
    #  "奏":"かなで",
    "何は": "なにが",
    "お父様": "おとうさま",
    "お兄様": "おにいさま",
    "何を": "なにを",
    "良い": "いい",
    "李衣菜": "りいな",
    "志希": "しき",
    "種": "たね",
    "方々": "かたがた",
    "颯": "はやて",
    "茄子さん": "かこさん",
    "茄子ちゃん": "かこちゃん",
    "涼ちゃん": "りょうちゃん",
    "涼さん": "りょうさん",
    "紗枝": "さえ",
    "文香": "ふみか",
    "私": "わたし",
    "周子": "しゅうこ",
    "イェ": "いえ",
    "可憐": "かれん",
    "加蓮": "かれん",
    "･": ".",
    #  "方の":"かたの",
    #  "気に":"きに",
    "唯さん": "ゆいさん",
    "唯ちゃん": "ゆいちゃん",
    "聖ちゃん": "ひじりちゃん",
    "他の": "ほかの",
    "他に": "ほかに",
    "一生懸命": "いっしょうけんめい",
    "楓さん": "かえでさん",
    "楓ちゃん": "かえでちゃん",
    "内から": "ないから",
    "の下で": "のしたで",

}

mapper = dict([

    ("仕方", "しかた"),
    ("明日", "あした"),
    ('私', "わたし"),
    ("従妹", "いとこ"),

    ("1人", "ひとり"),
    ("2人", "ふたり"),

    ("一期", "いちご"),
    ("一会", "いちえ"),

    ("♪", "！"),
    ("?", "？"),

    ("どんな方", "どんなかた"),
    ("ふたり暮らし", "ふたりぐらし"),

    ("新年", "しんねん"),
    ("来年", "らいねん"),
    ("去年", "きょねん"),
    ("壮年", "そうねん"),
    ("今年", "ことし"),

    ("昨年", "さくねん"),
    ("本年", "ほんねん"),
    ("平年", "へいねん"),
    ("閏年", "うるうどし"),
    ("初年", "しょねん"),
    ("少年", "しょうねん"),
    ("多年", "たねん"),
    ("青年", "せいねん"),
    ("中年", "ちゅうねん"),
    ("老年", "ろうねん"),
    ("成年", "せいねん"),
    ("幼年", "ようねん"),
    ("前年", "ぜんねん"),
    ("元年", "がんねん"),
    ("経年", "けいねん"),
    ("当年", "とうねん"),

    ("明年", "みょうねん"),
    ("歳年", "さいねん"),
    ("数年", "すうねん"),
    ("半年", "はんとし"),
    ("後年", "こうねん"),
    ("実年", "じつねん"),
    ("年年", "ねんねん"),
    ("連年", "れんねん"),
    ("暦年", "れきねん"),
    ("各年", "かくねん"),
    ("全年", "ぜんねん"),

    ("年を", "としを"),
    ("年が", "としが"),
    ("年も", "としも"),
    ("年は", "としは"),

    ("奏ちゃん", "かなでちゃん"),
    ("負けず嫌い", "まけずぎらい"),
    ("貴方", "あなた"),
    ("貴女", "あなた"),
    ("貴男", "あなた"),

    ("その節", "そのせつ"),

    ("何し", "なにし"),
    ("何する", "なにする"),

    ("心さん", "しんさん"),
    ("心ちゃん", "しんちゃん"),

    ("乃々", "のの"),

    ("身体の", "からだの"),
    ("身体が", "からだが"),
    ("身体を", "からだを"),
    ("身体は", "からだは"),
    ("身体に", "からだに"),
    ("正念場", "しょうねんば"),
    ("言う", "いう"),

    ("一回", "いっかい"),
    ("一曲", "いっきょく"),
    ("一日", "いちにち"),
    ("一言", "ひとこと"),
    ("一杯", "いっぱい"),

    ("方が", "ほうが"),
    ("縦輪城", "じゅうりんしろ"),
    ("深息", "しんそく"),
    ("家人", "かじん"),
    ("お返し", "おかえし"),
    ("化物語", "ばけものがたり"),
    ("阿良々木暦", "あららぎこよみ"),
    ("何より", "なにより")

])

# Merge all dictionaries into one
all_transformations = {**formal_to_informal, **formal_to_informal2, **formal_to_informal3, **mapper}


def apply_transformations(text, transformations=all_transformations):
    for key, value in transformations.items():
        text = text.replace(key, value)
    return text


def number_to_japanese(num):
    if not isinstance(num, int) or num < 0 or num > 9999:
        return "Invalid input"

    digits = ["", "いち", "に", "さん", "よん", "ご", "ろく", "なな", "はち", "きゅう"]
    tens = ["", "じゅう", "にじゅう", "さんじゅう", "よんじゅう", "ごじゅう", "ろくじゅう", "ななじゅう", "はちじゅう",
            "きゅうじゅう"]
    hundreds = ["", "ひゃく", "にひゃく", "さんびゃく", "よんひゃく", "ごひゃく", "ろっぴゃく", "ななひゃく",
                "はっぴゃく", "きゅうひゃく"]
    thousands = ["", "せん", "にせん", "さんぜん", "よんせん", "ごせん", "ろくせん", "ななせん", "はっせん",
                 "きゅうせん"]

    if num == 0:
        return "ゼロ"

    result = ""
    if num >= 1000:
        result += thousands[num // 1000]
        num %= 1000
    if num >= 100:
        result += hundreds[num // 100]
        num %= 100
    if num >= 10:
        result += tens[num // 10]
        num %= 10
    if num > 0:
        result += digits[num]

    return result


def convert_numbers_in_string(input_string):
    # Regular expression to find numbers in the string
    number_pattern = re.compile(r'\d+')

    # Function to replace numbers with their Japanese pronunciation
    def replace_with_japanese(match):
        num = int(match.group())
        return number_to_japanese(num)

    # Replace all occurrences of numbers in the string
    converted_string = number_pattern.sub(replace_with_japanese, input_string)
    return converted_string


roma_mapper = dict([

    ################################

    ("my", "mʲ"),
    ("by", "bʲ"),
    ("ny", "nʲ"),
    ("ry", "rʲ"),
    ("si", "sʲ"),
    ("ky", "kʲ"),
    ("gy", "gʲ"),
    ("dy", "dʲ"),
    ("di", "dʲ"),
    ("fi", "fʲ"),
    ("fy", "fʲ"),
    ("ch", "tɕ"),
    ("sh", "ɕ"),

    ################################

    ("a", "a"),
    ("i", "i"),
    ("u", "ɯ"),
    ("e", "e"),
    ("o", "o"),
    ("ka", "ka"),
    ("ki", "ki"),
    ("ku", "kɯ"),
    ("ke", "ke"),
    ("ko", "ko"),
    ("sa", "sa"),
    ("shi", "ɕi"),
    ("su", "sɯ"),
    ("se", "se"),
    ("so", "so"),
    ("ta", "ta"),
    ("chi", "tɕi"),
    ("tsu", "tsɯ"),
    ("te", "te"),
    ("to", "to"),
    ("na", "na"),
    ("ni", "ni"),
    ("nu", "nɯ"),
    ("ne", "ne"),
    ("no", "no"),
    ("ha", "ha"),
    ("hi", "çi"),
    ("fu", "ɸɯ"),
    ("he", "he"),
    ("ho", "ho"),
    ("ma", "ma"),
    ("mi", "mi"),
    ("mu", "mɯ"),
    ("me", "me"),
    ("mo", "mo"),
    ("ra", "ɽa"),
    ("ri", "ɽi"),
    ("ru", "ɽɯ"),
    ("re", "ɽe"),
    ("ro", "ɽo"),
    ("ga", "ga"),
    ("gi", "gi"),
    ("gu", "gɯ"),
    ("ge", "ge"),
    ("go", "go"),
    ("za", "za"),
    ("ji", "dʑi"),
    ("zu", "zɯ"),
    ("ze", "ze"),
    ("zo", "zo"),
    ("da", "da"),

    ("zu", "zɯ"),
    ("de", "de"),
    ("do", "do"),
    ("ba", "ba"),
    ("bi", "bi"),
    ("bu", "bɯ"),
    ("be", "be"),
    ("bo", "bo"),
    ("pa", "pa"),
    ("pi", "pi"),
    ("pu", "pɯ"),
    ("pe", "pe"),
    ("po", "po"),
    ("ya", "ja"),
    ("yu", "jɯ"),
    ("yo", "jo"),
    ("wa", "wa"),

    ("a", "a"),
    ("i", "i"),
    ("u", "ɯ"),
    ("e", "e"),
    ("o", "o"),
    ("wa", "wa"),
    ("o", "o"),

    ("wo", "o")])

nasal_sound = dict([
    # before m, p, b
    ("ɴm", "mm"),
    ("ɴb", "mb"),
    ("ɴp", "mp"),

    # before k, g
    ("ɴk", "ŋk"),
    ("ɴg", "ŋg"),

    # before t, d, n, s, z, ɽ
    ("ɴt", "nt"),
    ("ɴd", "nd"),
    ("ɴn", "nn"),
    ("ɴs", "ns"),
    ("ɴz", "nz"),
    ("ɴɽ", "nɽ"),

    ("ɴɲ", "ɲɲ"),

])


def Roma2IPA(text):
    orig = text

    for k, v in roma_mapper.items():
        text = text.replace(k, v)

    return text


def nasal_mapper(text):
    orig = text

    for k, v in nasal_sound.items():
        text = text.replace(k, v)

    return text


def alphabetreading(text):
    alphabet_dict = {"A": "エイ",
                     "B": "ビー",
                     "C": "シー",
                     "D": "ディー",
                     "E": "イー",
                     "F": "エフ",
                     "G": "ジー",
                     "H": "エイチ",
                     "I": "アイ",
                     "J": "ジェイ",
                     "K": "ケイ",
                     "L": "エル",
                     "M": "エム",
                     "N": "エヌ",
                     "O": "オー",
                     "P": "ピー",
                     "Q": "キュー",
                     "R": "アール",
                     "S": "エス",
                     "T": "ティー",
                     "U": "ユー",
                     "V": "ヴィー",
                     "W": "ダブリュー",
                     "X": "エックス",
                     "Y": "ワイ",
                     "Z": "ゼッド"}
    text = text.upper()
    text_ret = ""
    for t in text:
        if t in alphabet_dict:
            text_ret += alphabet_dict[t]
        else:
            text_ret += t
    return text_ret


import re
import cutlet

roma_mapper_plus_2 = {

    "bjo": 'bʲo',
    "rjo": "rʲo",
    "kjo": "kʲo",
    "kyu": "kʲu",

}


def replace_repeated_chars(input_string):
    result = []
    i = 0
    while i < len(input_string):
        if i + 1 < len(input_string) and input_string[i] == input_string[i + 1] and input_string[i] in 'aiueo':
            result.append(input_string[i] + 'ː')
            i += 2
        else:
            result.append(input_string[i])
            i += 1
    return ''.join(result)


def replace_chars_2(text, mapping=roma_mapper_plus_2):
    sorted_keys = sorted(mapping.keys(), key=len, reverse=True)

    pattern = '|'.join(re.escape(key) for key in sorted_keys)

    def replace(match):
        key = match.group(0)
        return mapping.get(key, key)

    return re.sub(pattern, replace, text)


def replace_tashdid_2(s):
    vowels = 'aiueoɯ0123456789.?!_。؟？！．．．＠@＃#＄$％%＾^＆&＊*（)(）_+=[「」]></\`~～―ー∺"'
    result = []

    i = 0
    while i < len(s):
        if i < len(s) - 2 and s[i].lower() == s[i + 2].lower() and s[i].lower() not in vowels and s[i + 1] == ' ':
            result.append('ʔ')
            result.append(s[i + 2])
            i += 3
        elif i < len(s) - 1 and s[i].lower() == s[i + 1].lower() and s[i].lower() not in vowels:
            result.append('ʔ')
            result.append(s[i + 1])
            i += 2
        else:
            result.append(s[i])
            i += 1

    return ''.join(result)


def replace_tashdid(input_string):
    result = []
    i = 0
    while i < len(input_string):
        if i + 1 < len(input_string) and input_string[i] == input_string[i + 1] and input_string[i] not in 'aiueo':
            result.append('ʔ')
            result.append(input_string[i])
            i += 2  # Skip the next character as it is already processed
        else:
            result.append(input_string[i])
            i += 1
    return ''.join(result)


def hira2ipa(text, roma_mapper=roma_mapper):
    keys_set = set(roma_mapper.keys())
    special_rule = ("n", "ɴ")

    transformed_text = []
    i = 0

    while i < len(text):
        if text[i] == special_rule[0]:
            if i + 1 == len(text) or text[i + 1] not in keys_set:
                transformed_text.append(special_rule[1])
            else:
                transformed_text.append(text[i])
        else:
            transformed_text.append(text[i])

        i += 1

    return ''.join(transformed_text)


k_mapper = dict([
    ("ゔぁ","ba"),
    ("ゔぃ","bi"),
    ("ゔぇ","be"),
    ("ゔぉ","bo"),
    ("ゔゃ","bʲa"),
    ("ゔゅ","bʲɯ"),
    ("ゔゃ","bʲa"),
    ("ゔょ","bʲo"),

    ("ゔ","bɯ"),

    ("あぁ"," aː"),
    ("いぃ"," iː"),
    ("いぇ"," je"),
    ("いゃ"," ja"),
    ("うぅ"," ɯː"),
    ("えぇ"," eː"),
    ("おぉ"," oː"),
    ("かぁ"," kaː"),
    ("きぃ"," kiː"),
    ("くぅ","kɯː"),
    ("くゃ","ka"),
    ("くゅ","kʲɯ"),
    ("くょ","kʲo"),
    ("けぇ","keː"),
    ("こぉ","koː"),
    ("がぁ","gaː"),
    ("ぎぃ","giː"),
    ("ぐぅ","gɯː"),
    ("ぐゃ","gʲa"),
    ("ぐゅ","gʲɯ"),
    ("ぐょ","gʲo"),
    ("げぇ","geː"),
    ("ごぉ","goː"),
    ("さぁ","saː"),
    ("しぃ","ɕiː"),
    ("すぅ","sɯː"),
    ("すゃ","sʲa"),
    ("すゅ","sʲɯ"),
    ("すょ","sʲo"),
    ("せぇ","seː"),
    ("そぉ","soː"),
    ("ざぁ","zaː"),
    ("じぃ","dʑiː"),
    ("ずぅ","zɯː"),
    ("ずゃ","zʲa"),
    ("ずゅ","zʲɯ"),
    ("ずょ","zʲo"),
    ("ぜぇ","zeː"),
    ("ぞぉ","zeː"),
    ("たぁ","taː"),
    ("ちぃ","tɕiː"),
    ("つぁ","tsa"),
    ("つぃ","tsi"),
    ("つぅ","tsɯː"),
    ("つゃ","tɕa"),
    ("つゅ","tɕɯ"),
    ("つょ","tɕo"),
    ("つぇ","tse"),
    ("つぉ","tso"),
    ("てぇ","teː"),
    ("とぉ","toː"),
    ("だぁ","daː"),
    ("ぢぃ","dʑiː"),
    ("づぅ","dɯː"),
    ("づゃ","zʲa"),
    ("づゅ","zʲɯ"),
    ("づょ","zʲo"),
    ("でぇ","deː"),
    ("どぉ","doː"),
    ("なぁ","naː"),
    ("にぃ","niː"),
    ("ぬぅ","nɯː"),
    ("ぬゃ","nʲa"),
    ("ぬゅ","nʲɯ"),
    ("ぬょ","nʲo"),
    ("ねぇ","neː"),
    ("のぉ","noː"),
    ("はぁ","haː"),
    ("ひぃ","çiː"),
    ("ふぅ","ɸɯː"),
    ("ふゃ","ɸʲa"),
    ("ふゅ","ɸʲɯ"),
    ("ふょ","ɸʲo"),
    ("へぇ","heː"),
    ("ほぉ","hoː"),
    ("ばぁ","baː"),
    ("びぃ","biː"),
    ("ぶぅ","bɯː"),
    ("ふゃ","ɸʲa"),
    ("ぶゅ","bʲɯ"),
    ("ふょ","ɸʲo"),
    ("べぇ","beː"),
    ("ぼぉ","boː"),
    ("ぱぁ","paː"),
    ("ぴぃ","piː"),
    ("ぷぅ","pɯː"),
    ("ぷゃ","pʲa"),
    ("ぷゅ","pʲɯ"),
    ("ぷょ","pʲo"),
    ("ぺぇ","peː"),
    ("ぽぉ","poː"),
    ("まぁ","maː"),
    ("みぃ","miː"),
    ("むぅ","mɯː"),
    ("むゃ","mʲa"),
    ("むゅ","mʲɯ"),
    ("むょ","mʲo"),
    ("めぇ","meː"),
    ("もぉ","moː"),
    ("やぁ","jaː"),
    ("ゆぅ","jɯː"),
    ("ゆゃ","jaː"),
    ("ゆゅ","jɯː"),
    ("ゆょ","joː"),
    ("よぉ","joː"),
    ("らぁ","ɽaː"),
    ("りぃ","ɽiː"),
    ("るぅ","ɽɯː"),
    ("るゃ","ɽʲa"),
    ("るゅ","ɽʲɯ"),
    ("るょ","ɽʲo"),
    ("れぇ","ɽeː"),
    ("ろぉ","ɽoː"),
    ("わぁ","ɯaː"),
    ("をぉ","oː"),

    ("う゛","bɯ"),
    ("でぃ","di"),
    ("でぇ","deː"),
    ("でゃ","dʲa"),
    ("でゅ","dʲɯ"),
    ("でょ","dʲo"),
    ("てぃ","ti"),
    ("てぇ","teː"),
    ("てゃ","tʲa"),
    ("てゅ","tʲɯ"),
    ("てょ","tʲo"),
    ("すぃ","si"),
    ("ずぁ","zɯa"),
    ("ずぃ","zi"),
    ("ずぅ","zɯ"),
    ("ずゃ","zʲa"),
    ("ずゅ","zʲɯ"),
    ("ずょ","zʲo"),
    ("ずぇ","ze"),
    ("ずぉ","zo"),
    ("きゃ","kʲa"),
    ("きゅ","kʲɯ"),
    ("きょ","kʲo"),
    ("しゃ","ɕʲa"),
    ("しゅ","ɕʲɯ"),
    ("しぇ","ɕʲe"),
    ("しょ","ɕʲo"),
    ("ちゃ","tɕa"),
    ("ちゅ","tɕɯ"),
    ("ちぇ","tɕe"),
    ("ちょ","tɕo"),
    ("とぅ","tɯ"),
    ("とゃ","tʲa"),
    ("とゅ","tʲɯ"),
    ("とょ","tʲo"),
    ("どぁ","doa"),
    ("どぅ","dɯ"),
    ("どゃ","dʲa"),
    ("どゅ","dʲɯ"),
    ("どょ","dʲo"),
    ("どぉ","doː"),
    ("にゃ","nʲa"),
    ("にゅ","nʲɯ"),
    ("にょ","nʲo"),
    ("ひゃ","çʲa"),
    ("ひゅ","çʲɯ"),
    ("ひょ","çʲo"),
    ("みゃ","mʲa"),
    ("みゅ","mʲɯ"),
    ("みょ","mʲo"),
    ("りゃ","ɽʲa"),
    ("りぇ","ɽʲe"),
    ("りゅ","ɽʲɯ"),
    ("りょ","ɽʲo"),
    ("ぎゃ","gʲa"),
    ("ぎゅ","gʲɯ"),
    ("ぎょ","gʲo"),
    ("ぢぇ","dʑe"),
    ("ぢゃ","dʑa"),
    ("ぢゅ","dʑɯ"),
    ("ぢょ","dʑo"),
    ("じぇ","dʑe"),
    ("じゃ","dʑa"),
    ("じゅ","dʑɯ"),
    ("じょ","dʑo"),
    ("びゃ","bʲa"),
    ("びゅ","bʲɯ"),
    ("びょ","bʲo"),
    ("ぴゃ","pʲa"),
    ("ぴゅ","pʲɯ"),
    ("ぴょ","pʲo"),
    ("うぁ","ɯa"),
    ("うぃ","ɯi"),
    ("うぇ","ɯe"),
    ("うぉ","ɯo"),
    ("うゃ","ɯʲa"),
    ("うゅ","ɯʲɯ"),
    ("うょ","ɯʲo"),
    ("ふぁ","ɸa"),
    ("ふぃ","ɸi"),
    ("ふぅ","ɸɯ"),
    ("ふゃ","ɸʲa"),
    ("ふゅ","ɸʲɯ"),
    ("ふょ","ɸʲo"),
    ("ふぇ","ɸe"),
    ("ふぉ","ɸo"),

    ("あ"," a"),
    ("い"," i"),
    ("う","ɯ"),
    ("え"," e"),
    ("お"," o"),
    ("か"," ka"),
    ("き"," ki"),
    ("く"," kɯ"),
    ("け"," ke"),
    ("こ"," ko"),
    ("さ"," sa"),
    ("し"," ɕi"),
    ("す"," sɯ"),
    ("せ"," se"),
    ("そ"," so"),
    ("た"," ta"),
    ("ち"," tɕi"),
    ("つ"," tsɯ"),
    ("て"," te"),
    ("と"," to"),
    ("な"," na"),
    ("に"," ni"),
    ("ぬ"," nɯ"),
    ("ね"," ne"),
    ("の"," no"),
    ("は"," ha"),
    ("ひ"," çi"),
    ("ふ"," ɸɯ"),
    ("へ"," he"),
    ("ほ"," ho"),
    ("ま"," ma"),
    ("み"," mi"),
    ("む"," mɯ"),
    ("め"," me"),
    ("も"," mo"),
    ("ら"," ɽa"),
    ("り"," ɽi"),
    ("る"," ɽɯ"),
    ("れ"," ɽe"),
    ("ろ"," ɽo"),
    ("が"," ga"),
    ("ぎ"," gi"),
    ("ぐ"," gɯ"),
    ("げ"," ge"),
    ("ご"," go"),
    ("ざ"," za"),
    ("じ"," dʑi"),
    ("ず"," zɯ"),
    ("ぜ"," ze"),
    ("ぞ"," zo"),
    ("だ"," da"),
    ("ぢ"," dʑi"),
    ("づ"," zɯ"),
    ("で"," de"),
    ("ど"," do"),
    ("ば"," ba"),
    ("び"," bi"),
    ("ぶ"," bɯ"),
    ("べ"," be"),
    ("ぼ"," bo"),
    ("ぱ"," pa"),
    ("ぴ"," pi"),
    ("ぷ"," pɯ"),
    ("ぺ"," pe"),
    ("ぽ"," po"),
    ("や"," ja"),
    ("ゆ"," jɯ"),
    ("よ"," jo"),
    ("わ"," wa"),
    ("ゐ"," i"),
    ("ゑ"," e"),
    ("ん"," ɴ"),
    ("っ"," ʔ"),
    ("ー"," ː"),

    ("ぁ"," a"),
    ("ぃ"," i"),
    ("ぅ"," ɯ"),
    ("ぇ"," e"),
    ("ぉ"," o"),
    ("ゎ"," ɯa"),
    ("ぉ"," o"),
    ("っ","?"),

    ("を","o")

])


def post_fix(text):
    orig = text

    for k, v in k_mapper.items():
        text = text.replace(k, v)

    return text


sym_ws = dict([

    ("$ ", "dorɯ"),
    ("＄ ", "dorɯ"),

    ("〇 ", "marɯ"),
    ("¥ ", "eɴ"),

    ("# ", "haʔɕɯ tagɯ"),
    ("＃ ", "haʔɕɯ tagɯ"),

    ("& ", "ando"),
    ("＆ ", "ando"),

    ("% ", "paːsento"),
    ("％ ", "paːsento"),

    ("@ ", "aʔto saiɴ"),
    ("＠ ", "aʔto saiɴ")

])


def random_sym_fix(text):  # with space
    orig = text

    for k, v in sym_ws.items():
        text = text.replace(k, f" {v} ")

    return text


sym_ns = dict([

    ("$", "dorɯ"),
    ("＄", "dorɯ"),

    ("〇", "marɯ"),
    ("¥", "eɴ"),

    ("#", "haʔɕɯ tagɯ"),
    ("＃", "haʔɕɯ tagɯ"),

    ("&", "ando"),
    ("＆", "ando"),

    ("%", "paːsento"),
    ("％", "paːsento"),

    ("@", "aʔto saiɴ"),
    ("＠", "aʔto saiɴ"),

    ("～", "—"),
    ("kʲɯɯdʑɯɯkʲɯɯ.kʲɯɯdʑɯɯ", "kʲɯɯdʑɯɯ kʲɯɯ teɴ kʲɯɯdʑɯɯ")

])


def random_sym_fix_no_space(text):
    orig = text

    for k, v in sym_ns.items():
        text = text.replace(k, f" {v} ")

    return text


spaces = dict([

    ("ɯ ɴ", "ɯɴ"),
    ("na ɴ ", "naɴ "),
    (" mina ", " miɴna "),
    ("ko ɴ ni tɕi ha", "konnitɕiwa"),
    ("ha i", "hai"),
    ("boɯtɕama", "boʔtɕama"),
    ("i eːi", "ieːi"),
    ("taiɕɯtsɯdʑoɯ", "taiɕitsɯdʑoɯ"),
    ("soɴna ka ze ni", "soɴna fɯɯ ni"),
    (" i e ", "ke "),
    ("�", ""),
    ("×", " batsɯ "),
    ("se ka ɯndo", "sekaɯndo"),
    ("i i", "iː"),
    ("i tɕi", "itɕi"),
    ("ka i", "kai"),
    ("naɴ ga", "nani ga"),
    ("i eː i", "ieːi"),

    ("naɴ koɽe", "nani koɽe"),
    ("naɴ soɽe", "nani soɽe"),
    (" ɕeɴ ", " seɴ "),

    # ("konna","koɴna"),
    # ("sonna"," soɴna "),
    # ("anna","aɴna"),
    # ("nn","ɴn"),

    ("en ", "eɴ "),
    ("in ", "iɴ "),
    ("an ", "aɴ "),
    ("on ", "oɴ "),
    ("ɯn ", "ɯɴ "),
    # ("nd","ɴd"),

    ("koɴd o", "kondo"),
    ("ko ɴ d o", "kondo"),
    ("ko ɴ do", "kondo"),

    ("oanitɕaɴ", "oniːtɕaɴ"),
    ("oanisaɴ", "oniːsaɴ"),
    ("oanisama", "oniːsama"),
    ("hoːmɯrɯɴɯ", "hoːmɯrɯːmɯ"),
    ("so ɴ na ", "sonna"),
    ("  sonna  ", " sonna "),
    ("  konna  ", " konna "),
    ("ko ɴ na ", "konna"),
    (" ko to  ", " koto "),
    ("edʑdʑi", "eʔtɕi"),
    (" edʑdʑ ", " eʔtɕi "),
    (" dʑdʑ ", " dʑiːdʑiː "),
    ("secɯnd", "sekaɯndo"),

    ("ɴɯ", "nɯ"),
    ("ɴe", "ne"),
    ("ɴo", "no"),
    ("ɴa", "na"),
    ("ɴi", "ni"),
    ("ɴʲ", "nʲ"),

    ("hotond o", "hotondo"),
    ("hakoɴd e", "hakoɴde"),
    ("gakɯtɕi ɽi", "gaʔtɕiɽi "),

    (" ʔ", "ʔ"),
    ("ʔ ", "ʔ"),

    ("-", "ː"),
    ("- ", "ː"),
    ("--", "~ː"),
    ("~", "—"),
    ("、", ","),

    (" ː", "ː"),
    ('ka nade', "kanade"),

    ("ohahasaɴ", "okaːsaɴ"),
    ("　", " "),
    ("viː", "bɯiː"),
    ("ːː", "ː—"),

    ("d ʑ", "dʑ"),
    ("d a", "da"),
    ("d e", "de"),
    ("d o", "do"),
    ("d ɯ", "dɯ"),

    ("niːɕiki", "ni iɕiki"),
    ("anitɕaɴ", "niːtɕaɴ"),
    ("daiːtɕi", "dai itɕi"),
    ("niːta", "ni ita"),
    ("niːrɯ", "ni irɯ"),
    ("a—", "aː"),
    ("waːis", "wa ais"),
    ("waːiɕ", "wa aiɕ"),
    ("aːt", "a at"),
    ("waːʔ", "wa aʔ"),

    ("naɴ sono", "nani sono"),
    ("naɴ kono", "nani kono"),
    ("naɴ ano", "nani ano"),  # Cutlet please fix your shit
    (" niːtaɽa", " ni itaɽa"),
    ("doɽamaɕiːd", "doɽama ɕiːdʲi"),
    ("aɴ ta", "anta"),
    ("aɴta", "anta"),
    ("naniːʔteɴ", "nani iʔteɴ"),
    ("niːkite", "ni ikite"),
    ("niːʔ", "ni iʔ"),
    ("niːɯ", "ni iɯ"),
    ("niːw", "ni iw"),
    ("niːkɯ", "ni ikɯ"),
    ("de—", "de e"),
    ("aːj", "aː aj"),
    ("aːɽ", "a aɽ"),
    ("aːr", "a ar"),
    ("gaːn", "ga an"),
    ("ɕiːk ", "ɕi ik"),
    ("ɕijoː neɴ", "ɕoɯneɴ"),
    ("aːna", "a ana"),

])


def random_space_fix(text):
    orig = text

    for k, v in spaces.items():
        text = text.replace(k, v)

    return text


def replace_repeating_a(output):
    # Define patterns and their replacements
    patterns = [
        (r'(aː)\s*\1+\s*', r'\1~'),  # Replace repeating "aː" with "aː~~"
        (r'(aːa)\s*aː', r'\1~'),  # Replace "aːa aː" with "aː~~"
        (r'aːa', r'aː~'),  # Replace "aːa" with "aː~"
        (r'naː\s*aː', r'naː~'),  # Replace "naː aː" with "naː~"
        (r'(oː)\s*\1+\s*', r'\1~'),  # Replace repeating "oː" with "oː~~"
        (r'(oːo)\s*oː', r'\1~'),  # Replace "oːo oː" with "oː~~"
        (r'oːo', r'oː~'),  # Replace "oːo" with "oː~"
        (r'(eː)\s*\1+\s*', r'\1~'),
        (r'(e)\s*\1+\s*', r'\1~'),
        (r'(eːe)\s*eː', r'\1~'),
        (r'eːe', r'eː~'),
        (r'neː\s*eː', r'neː~'),
    ]

    # Apply each pattern to the output
    for pattern, replacement in patterns:
        output = re.sub(pattern, replacement, output)

    return output



class Phonemizer:
    def __init__(self):
        self.katsu = cutlet.Cutlet(ensure_ascii=False)
        self.katsu.use_foreign_spelling = False

    def _process_japanese_text(self, ml):
        # Check for small characters and replace them
        if any(char in ml for char in "ぁぃぅぇぉ"):
            ml = ml.replace("ぁ", "あ")
            ml = ml.replace("ぃ", "い")
            ml = ml.replace("ぅ", "う")
            ml = ml.replace("ぇ", "え")
            ml = ml.replace("ぉ", "お")

        # Initialize Cutlet for romaji conversion

        # Convert to romaji and apply transformations
        # output = katsu.romaji(ml, capitalize=False).lower()

        output = self.katsu.romaji(apply_transformations(alphabetreading(ml)), capitalize=False).lower()

        # Replace specific romaji sequences
        if 'j' in output:
            output = output.replace('j', "dʑ")
        if 'tt' in output:
            output = output.replace('tt', "ʔt")
        if 't t' in output:
            output = output.replace('t t', "ʔt")
        if ' ʔt' in output:
            output = output.replace(' ʔt', "ʔt")
        if 'ssh' in output:
            output = output.replace('ssh', "ɕɕ")

        # Convert romaji to IPA
        output = Roma2IPA(convert_numbers_in_string(output))

        output = hira2ipa(output)

        # Apply additional transformations
        output = replace_chars_2(output)
        output = replace_repeated_chars(replace_tashdid_2(output))
        output = nasal_mapper(output)

        # Final adjustments
        if " ɴ" in output:
            output = output.replace(" ɴ", "ɴ")

        if ' neɽitai ' in output:
            output = output.replace(' neɽitai ', "naɽitai")

        if 'harɯdʑisama' in output:
            output = output.replace('harɯdʑisama', "arɯdʑisama")

        if "ki ni ɕinai" in output:
            output = re.sub(r'(?<!\s)ki ni ɕinai', r' ki ni ɕinai', output)

        if 'ʔt' in output:
            output = re.sub(r'(?<!\s)ʔt', r'ʔt', output)

        if 'de aɽoɯ' in output:
            output = re.sub(r'(?<!\s)de aɽoɯ', r' de aɽoɯ', output)

        return output.lstrip()

    # def replace_repeating_patterns(text):
    #     def replace_repeats(match):
    #         pattern = match.group(1)
    #         if len(match.group(0)) // len(pattern) >= 3:
    #             return pattern + "~~~"
    #         return match.group(0)

    #     # Pattern for space-separated repeats
    #     pattern1 = r'((?:\S+\s+){1,5}?)(?:\1){2,}'
    #     # Pattern for continuous repeats without spaces
    #     pattern2 = r'(.+?)\1{2,}'

    #     text = re.sub(pattern1, replace_repeats, text)
    #     text = re.sub(pattern2, replace_repeats, text)
    #     return text

    def phonemize(self, text):
        # if "っ" in text:
        #     text = text.replace("っ","ʔ")

        output = post_fix(self._process_japanese_text(text))
        # output = text

        if " ɴ" in output:
            output = output.replace(" ɴ", "ɴ")
        if "y" in output:
            output = output.replace("y", "j")
        if "ɯa" in output:
            output = output.replace("ɯa", "wa")

        if "a aː" in output:
            output = output.replace("a aː", "a~")
        if "a a" in output:
            output = output.replace("a a", "a~")

        output = replace_repeating_a((output))
        output = re.sub(r'\s+~', '~', output)

        if "oː~o oː~ o" in output:
            output = output.replace("oː~o oː~ o", "oː~~~~~~")
        if "aː~aː" in output:
            output = output.replace("aː~aː", "aː~~~")
        if "oɴ naː" in output:
            output = output.replace("oɴ naː", "onnaː")
        if "aː~~ aː" in output:
            output = output.replace("aː~~ aː", "aː~~~~")
        if "oː~o" in output:
            output = output.replace("oː~o", "oː~~")
        if "oː~~o o" in output:
            output = output.replace("oː~~o o", "oː~~~~")  # yeah I'm too tired to learn regex how did you know

        output = random_space_fix(output)
        output = random_sym_fix(
            output)  # fixing some symbols, if they have a specific white space such as miku& sakura -> miku ando sakura
        output = random_sym_fix_no_space(
            output)  # same as above but for those without white space such as miku&sakura -> miku ando sakura
        # if "ɯ" in output:
        #     output = output.replace("ɯ","U")ｓｓ
        # if "ʔ" in output:
        #     output = output.replace("ʔ","!")

        if text.endswith(" "):
            output += " "

        return output.lstrip()



import os
import re

import nltk
from openphonemizer import OpenPhonemizer

from sentence_transformers import SentenceTransformer
from sklearn.metrics.pairwise import cosine_similarity

from pywsd.lesk import simple_lesk

nltk.download("wordnet", quiet=True)

from nltk.corpus import wordnet

os.environ['PYTHONIOENCODING'] = 'utf-8'

import epitran

# Initialize Epitran for IPA transcription
# Enforce UTF-8 globally
os.environ['PYTHONIOENCODING'] = 'utf-8'

try:
    epi = epitran.Epitran('eng-Latn')
except UnicodeDecodeError:
    raise OSError("epitran could not be loaded. if you're on windows, in control panel > region, "
                  "Check Beta: Use Unicode UTF-8 for worldwide language support")

from nltk.tokenize import word_tokenize

general = {
    # Basic contractions and common words
    "y'all": "jɔːl"
}

# Proper names and common mispronunciations
proper_names = {
    "Amazon": "ˈæməˌzɒn",
    "Microsoft": "ˈmaɪkrəˌsɒft",
    "Spotify": "ˈspɒtɪfaɪ",
    "Facebook": "ˈfeɪsˌbʊk",
    "Twitter": "ˈtwɪtər",
    "YouTube": "ˈjuːˌtjuːb",
    "Instagram": "ˈɪnstəˌɡræm",
    "Samsung": "ˈsæmˌsʌŋ",
    "Apple": "ˈæpəl",
    "Adobe": "əˈdoʊbi",
    "Beyoncé": "biˈjɒnseɪ",
    "Rihanna": "riˈɑːnə",
    "Kanye": "ˈkɑːnjeɪ",
    "J.K. Rowling": "ˌdʒeɪ.keɪ ˈroʊlɪŋ",
    "Harry Potter": "ˈhæri ˈpɒtər",
    "Marvel": "ˈmɑrvəl",
    "DC": "diː siː",
    "Pokemon": "ˈpoʊkɪmɒn",
    "Netflix": "ˈnɛtflɪks",
    "Siri": "ˈsɪri",
    "Alexa": "əˈlɛksə",
    "Tesla": "ˈtɛslə",
    "Quora": "ˈkwɔːrə",
    "Wikipedia": "ˌwɪkɪˈpiːdiə",
    "NVIDIA": "ɛnˈvɪdiə",
    "Snapchat": "ˈsnæpˌtʃæt",
    "LinkedIn": "ˈlɪŋktɪn",
    "Zoom": "zuːm",
    "Twitch": "twɪtʃ",
    "Kombucha": "kəmˈbuːtʃə",
    "Chia": "ˈtʃiːə",
    "Yelp": "jɛlp",
    "TikTok": "tɪkˈtɒk",
    "Duolingo": "ˌdjuːəˈlɪŋɡoʊ",
    "Coca-Cola": "ˈkoʊkəˌkoʊlə",
    "Pepsi": "ˈpɛpsi",
    "Starbucks": "ˈstɑrbʌks",
    "Walmart": "ˈwɔːlmɑːrt",
    "IKEA": "aɪˈkiːə",
    "Uber": "ˈjuːbər",
    "Lyft": "lɪft",
    "KFC": "keɪ ɛf ˈsiː",
    "NBA": "ɛn biː eɪ",
    "NFL": "ɛn ɛf ɛl",
    "FIFA": "ˈfiːfə",
    "NHL": "ɛn eɪtʃ ɛl",
    "Reddit": "ˈrɛdɪt",
    "Tinder": "ˈtɪndər",
    "WordPress": "ˈwɜrdprɛs",
}

# Common mispronunciations
common_mispronunciations = {
    "meme": "miːm",
    "pasta": "ˈpɑːstə",
    "quinoa": "ˈkiːnwɑː",
    "sriracha": "sɪˈrɑːtʃə",
    "coup": "kuː",
    "genre": "ˈʒɒnrə",
    "cliché": "kliːˈʃeɪ",
    "façade": "fəˈsɑːd",
    "entrepreneur": "ˌɒntrəprəˈnɜːr",
    "ballet": "bæˈleɪ",
    "jalapeño": "ˌhæləˈpeɪnjoʊ",
    "caramel": "ˈkærəˌmɛl",
    "vaccine": "vækˈsiːn",
    "herb": "hɜːrb",  # (often mispronounced as 'urb')
}

innacurate_from_phonemizer = {
    "british": "ˈbrɪt.ɪʃ"
}

# Combine both dictionaries
manual_phonemizations = {**general, **proper_names, **common_mispronunciations, **innacurate_from_phonemizer}

model = SentenceTransformer('all-MiniLM-L6-v2')

word_definitions = {
    "lead": {
        "to guide or direct": "liːd",
        "a type of metal": "lɛd"
    },
    "tear": {
        "separate or cause to separate abruptly": "tɛər",
        "fill with tears or shed tears": "tɪər"
    },
    "read": {
        "to look at and comprehend written words": "riːd",
        "past tense of read": "rɛd"
    },
    "wind": {
        "moving air": "wɪnd",
        "to twist or coil": "wɪnd"
    },
    "row": {
        "a linear arrangement of things": "roʊ",
        "to propel a boat": "raʊ"
    },
    "live": {
        "to be alive": "lɪv",
        "happening in real time": "laɪv"
    },
    "close": {
        "to shut something": "kloʊs",
        "near": "kloʊs"
    },
    "bass": {
        "a type of fish": "beɪs",
        "low-frequency sound or voice": "bæs"
    }
}

### ^^^ PLACEHOLDER UNTIL MANUAL DICT CREATED

def get_most_similar_definition(word, query):
    if word not in word_definitions:
        return "", word

    # Get the definitions of the word
    definitions = word_definitions[word]

    # Encode the query sentence and definitions using the model
    if query not in definitions:
        query_embedding = model.encode([query])
        definition_embeddings = model.encode(list(definitions.keys()))

        # Calculate cosine similarity between the query and the definitions
        similarities = cosine_similarity(query_embedding, definition_embeddings)

        # Find the index of the most similar definition
        most_similar_index = similarities.argmax()

        # Return the most similar definition
        most_similar_definition = list(definitions.keys())[most_similar_index]
        return most_similar_definition, definitions[most_similar_definition]
    else:
        return query, definitions[query]

# Function to check if a word is a homonym
def is_homonym(word):
    synsets = wordnet.synsets(word)

    filtered_sysnets = []

    for synset in synsets:
        if word == synset.name().split(".")[0].lower():
            filtered_sysnets.append(synset)

    return len(filtered_sysnets) > 1


# Function to generate pronunciation dictionary for homonyms
def generate_pronunciation_dict(word):
    synsets = wordnet.synsets(word)
    pronunciation_dict = {}

    for synset in synsets:
        if synset.name().split(".")[0].lower() == word:
            definition = synset.definition()
            # Get the lemma name for transcription
            lemma_name = synset.lemmas()[0].name().replace("_", " ")
            # Generate IPA pronunciation using Epitran
            _, ipa_pronunciation = get_most_similar_definition(word, definition)
            pronunciation_dict[definition] = ipa_pronunciation

    return pronunciation_dict


def replace_homonyms(text):
    verbose = False

    # Use regex to split text while keeping the delimiters
    tokens = re.findall(r'\S+|\s+', text)
    result = tokens.copy()

    # Keep track of homonym indices to process each separately
    homonym_indices = []

    # First pass: identify homonym indices
    for i, token in enumerate(tokens):
        if not token.isspace():
            word_lower = token.lower()
            if is_homonym(word_lower):
                homonym_indices.append(i)

    # Process each homonym separately
    for index in homonym_indices:
        # Create a context window around the current word
        context_start = max(0, index - 3)
        context_end = min(len(tokens), index + 5)

        # Extract context
        context_tokens = tokens[context_start:context_end]
        context = ''.join(context_tokens).lower()

        # Focus on the specific word
        current_word = tokens[index].lower()

        # Disambiguate meaning using Lesk with the specific context
        sense = simple_lesk(context, current_word)

        if sense:
            meaning = sense.definition()
            # Generate pronunciation dictionary for the word
            pronunciation_dict = generate_pronunciation_dict(current_word)

            # Find pronunciation for the matching meaning
            pronunciation = pronunciation_dict.get(meaning, current_word)

            # Verbose output if requested
            if verbose:
                # Prepare sliding context window with padding
                context_display = ''.join(context_tokens)
                context_width = 20  # Fixed width for consistent display
                context_padded = (context_display.center(context_width))[:context_width]

                # Prepare meaning with padding/trimming
                meaning_width = 50  # Fixed width for consistent display
                meaning_padded = (meaning[:meaning_width].center(meaning_width))[:meaning_width]

                # Print in a formatted way
                print(f"[{context_padded}] - {meaning_padded}\r")

            # Replace the token with phoneme representation
            result[index] = f"<phoneme>{pronunciation}</phoneme>"

    if verbose:
        print("")
    return ''.join(result)


class Phonemizer:
    def __init__(self, manual_fixes=None, allow_heteronyms=True, stress=False): # temporarily allow heteronyms until we fill the dictionary
        if manual_fixes is None:
            manual_fixes = manual_phonemizations
        self.phonemizer = OpenPhonemizer()

        # Dictionary of manual phonemizations
        self.manual_phonemizations = manual_fixes
        self.allow_heteronyms=allow_heteronyms
        self.stress = stress

        # Post-processing filters
        self.manual_filters = {
            " . . . ": "... ",
            " . ": ". "
        }

        # Regex to detect text wrapped in <phoneme> tags
        self.phoneme_tag_pattern = re.compile(r"<phoneme>(.*?)</phoneme>")

    def preprocess(self, text):
        if not self.allow_heteronyms:
            text = replace_homonyms(text)

        # Replace words in the text with their manual phonemizations wrapped in <phoneme> tags
        for word, ipa in self.manual_phonemizations.items():
            text = re.sub(rf"\b{word}\b", f"<phoneme>{ipa}</phoneme>", text, flags=re.IGNORECASE)
        return text

    def postprocess(self, text):
        # Remove the <phoneme> tags but retain the IPA within them, preserving spaces
        if not self.stress:
            text = re.sub("ˈ", "", text)
            text = re.sub("\u02C8", "", text) # double check

        return self.phoneme_tag_pattern.sub(r"\1", text)

    def phonemize(self, text):
        # Preprocess the text for manual phonemizations
        preprocessed_text = self.preprocess(text)

        result = []
        in_quotes = False
        current_segment = ""

        i = 0
        while i < len(preprocessed_text):
            # Check for phoneme tags
            phoneme_match = self.phoneme_tag_pattern.match(preprocessed_text, i)
            if phoneme_match:
                # Append the phoneme tag content and preserve spaces before and after
                if current_segment:
                    result.append(self.phonemizer(current_segment))
                    current_segment = ""

                result.append(phoneme_match.group(1))  # Add the IPA content directly
                i = phoneme_match.end()
                continue

            char = preprocessed_text[i]

            if char == '"':
                if current_segment:
                    if not in_quotes:
                        processed_segment = self.phonemizer(current_segment)
                    else:
                        processed_segment = f'{self.phonemizer(current_segment)}'
                    result.append(processed_segment)
                    current_segment = ""

                result.append(char)
                in_quotes = not in_quotes
            else:
                current_segment += char

            i += 1

        # Process any remaining text
        if current_segment:
            if not in_quotes:
                processed_segment = self.phonemizer(current_segment)
            else:
                processed_segment = f'"{self.phonemizer(current_segment)}"'
            result.append(processed_segment)

        phonemized_text = ''.join(result)

        # Apply manual filters
        for filter, item in self.manual_filters.items():
            phonemized_text = phonemized_text.replace(filter, item)

        # Post-process to remove phoneme tags
        final_text = self.postprocess(phonemized_text)

        return final_text


if __name__ == "__main__":
    phonem = Phonemizer()
    test_text = "i tear the paper apart, water welling up in my eyes, a tear in my eye"
    print(f"Original: {test_text}")
    print(f"Phonemized: {phonem.phonemize(test_text)}")



from ruphon import RUPhon
from ruaccent import RUAccent
import torch

if torch.cuda.is_available():
    device = "cuda"
else:
    device = "cpu"

class Phonemizer:
    def __init__(self, working_path=None, stress=False):
        self.stress = stress

        self.phonemizer = RUPhon()
        self.phonemizer = self.phonemizer.load("small", workdir=working_path, device=device)

        self.accentizer = RUAccent()
        self.accentizer.load(omograph_model_size='turbo3', use_dictionary=True, tiny_mode=False)

    def phonemize(self, text):
        accented_text = self.accentizer.process_all(text)

        result = self.phonemizer.phonemize(accented_text, put_stress=self.stress, stress_symbol="'")

        return result


from pythainlp.transliterate import transliterate
import warnings

class Phonemizer:
    def __init__(self):
        """
        The base thai phonemizer
        """

        warnings.warn("The thai phonemizer has been loaded, unfortunately we are unable "
                      "to determine the accuracy of this transcript, use at your own risk")

    @staticmethod
    def phonemize(text):
        print(text)
        result = transliterate(text, engine="thaig2p")
        result = (result.replace(" . ", "-.-")
                  .replace("  ", "-||-")
                  .replace(" ", "")
                  .replace("-||-", " ")
                  .replace("-.-", ". "))
        return result



import warnings
from termcolor import colored
from .phonemizers import english, japanese, mandarin, russian, thai
from .langtokenizers.multicoded import Tokenizer, LANGUAGE_COLORS
import re


class Phonemizer:
    """
    A class for phonemizing text in multiple languages,
    """

    def __init__(self, working_path=None, stress=False):
        """
        Initialize the Phonemizer.

        :param working_path: Optional path for working directory
        :param stress: Optional toggle for stress, for phonemisers that support it
        """
        self.working_path = working_path
        self.stress = stress
        self._phonemizers = {}
        self.Tokenizer = Tokenizer()

    def pretty_print(self, phonemized_dict: list):
        """
        Print the phonemized text with colors based on language.

        :param phonemized_dict: A list of dictionaries containing phonemized text segments and their languages
        """
        for segment in phonemized_dict:
            text = segment['text']
            lang = segment['lang']

            pattern = r'<\?\?>(.*?)</\?\?>'

            # Determine the color based on the language
            color = LANGUAGE_COLORS[lang]

            # Replace the tagged content with colored content
            def replace_with_color(match):
                return '\033[1m' + f"{match.group(1)}" + '\033[0m'

            # Substitute the pattern with colored text
            colored_text = re.sub(pattern, replace_with_color, text)

            # Print the text with the corresponding color
            print(colored(colored_text, color), end='')
        print("")

    def get_phonemizer(self, lang):
        """
        Get or create a phonemizer for the specified language.

        :param lang: Language code (e.g., 'en', 'ja', 'zh', 'ru')
        :return: A phonemizer instance for the specified language, or None if not supported
        """
        if lang not in self._phonemizers:
            if lang == 'en':
                self._phonemizers[lang] = english.Phonemizer(stress=self.stress)
            elif lang == 'ja':
                self._phonemizers[lang] = japanese.Phonemizer()
            elif lang == 'zh':
                self._phonemizers[lang] = mandarin.Phonemizer()
            elif lang == 'cy': # cyrillic treated as russian
                self._phonemizers[lang] = russian.Phonemizer(working_path=self.working_path)
            elif lang == 'th':
                self._phonemizers[lang] = thai.Phonemizer()
        return self._phonemizers.get(lang)

    def seperate_languages(self, text):
        """
        Separate the input text into segments based on language tags.

        :param text: Input text with language tags
        :return: A list of dictionaries containing text segments and their languages
        """
        text = self.Tokenizer.tokenize(text)

        pattern = r'(<(\w+)>(.*?)</\2>)|([^<]+)'
        matches = re.findall(pattern, text)

        result = []
        current_item = {"text": "", "lang": None}

        for match in matches:
            if match[1]:  # Tagged content
                lang, content = match[1], match[2]
                if current_item["lang"] != lang:
                    if current_item["text"]:
                        result.append(current_item)
                    current_item = {"text": content, "lang": lang}
                else:
                    current_item["text"] += content
            else:  # Untagged content (punctuation or spaces)
                untagged_content = match[3]
                if current_item["text"]:
                    current_item["text"] += untagged_content
                else:
                    result.append({"text": untagged_content, "lang": "untagged"})

        if current_item["text"]:
            result.append(current_item)

        return result

    def phonemize_for_language(self, text, lang):
        """
        Phonemize the given text for a specific language.

        :param text: The plaintext to phonemize
        :param lang: The language ID for phonemization
        :return: Phonemized text, or original text wrapped in <??> tags if language is not supported
        """
        if lang != "phoneme":
            phonemizer = self.get_phonemizer(lang)
            if phonemizer:
                return phonemizer.phonemize(text)
            return f"<??>{text}</??>"  # Return original text if no phonemizer available
        else:
            return text

    def phonemize(self, input_text, output_dict=False):
        """
        Phonemize the input text, handling multiple languages including CJK.

        :param input_text: The input text to phonemize
        :param output_dict: If True, return a list of dictionaries with text and language; if False, return a single string
        :return: Phonemized text as a string or list of dictionaries
        """
        separated = self.seperate_languages(input_text)
        result = []

        for item in separated:
            # if self.Tokenizer.detect_japanese_korean_chinese(item["text"]) != "??":
            #     result.append(self._process_cjk_segment(item))
            # else:
            result.append(item)


        phonemized_result = []
        for item in result:
            phonemized_text = self.phonemize_for_language(item['text'], item['lang'])
            if output_dict:
                lang = item["lang"] if "??" not in phonemized_text else "??"
                phonemized_result.append({"text": phonemized_text, "lang": lang})
            else:
                phonemized_result.append(phonemized_text)

        if not output_dict:
            fin = ''.join(phonemized_result)
            if "<??>" in fin:
                warnings.warn(
                    "Your output contains unsupported languages, "
                    "<??> tags have been added to allow for manual filtering")
            return fin
        else:
            return phonemized_result

    def _process_cjk_segment(self, item):
        """
        Process a CJK (Chinese, Japanese, Korean) text segment.

        :param item: A dictionary containing the text segment and its language
        :return: A list of dictionaries with processed CJK segments and their detected languages
        """
        processed_segments = []
        segmentsCJKog = self.Tokenizer.split_non_cjk_in_segment(item["text"])

        for CJKog in segmentsCJKog:
            phonemized_text = self.phonemize_for_language(CJKog, item['lang'])
            segmentsCJK = self.Tokenizer.split_non_cjk_in_segment(phonemized_text)

            for CJK in segmentsCJK:
                CJKLang = self.Tokenizer.detect_japanese_korean_chinese(CJK)
                if CJKLang != "??":
                    processed_segments.append({"text": CJK, "lang": CJKLang})

                    remaining = str(CJKog.split(CJK, 1)[-1])
                    remaining = remaining.replace(CJK.strip(), "")
                    if remaining:
                        remaining_lang = self.Tokenizer.detect_japanese_korean_chinese(remaining)
                        processed_segments.append({"text": remaining, "lang": remaining_lang})

        return processed_segments


if __name__ == "__main__":
    input_text = "hello, 你好は中国語でこんにちはと言う意味をしています。مرحبا! Привет! नमस्ते!"
    engine = Phonemizer()
    from time import time

    start = time()
    output = engine.phonemize(input_text, output_dict=True)
    end = time()
    print(input_text)
    engine.pretty_print(output)
    print(f"Took - First: {end - start}")

    start = time()
    output = engine.phonemize(input_text, output_dict=True)
    end = time()
    print(input_text)
    engine.pretty_print(output)
    print(f"Took - Instantiated: {end - start}")